#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

__version__ = '0.3.10'
__source__ = "https://gitlab.com/jsmetana/mcdbot"
__issues__ = "https://gitlab.com/jsmetana/mcdbot/-/issues"



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

import binascii
import hashlib


def get_offline_player_uuid(player: str):
    uuid = bytearray(hashlib.md5(b"OfflinePlayer:" + player.encode()).digest())
    uuid[6] = uuid[6] & 0x0f | 0x30
    uuid[8] = uuid[8] & 0x3f | 0x80
    uuid = binascii.hexlify(uuid).decode()
    return format_uuid(uuid)


def format_uuid(uuid: str):
    return '-'.join([uuid[0:8], uuid[8:12], uuid[12:16], uuid[16:20], uuid[20:32]])



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.errors.mcdbot_error import McdbotError
from mcdbot.config import Config
from mcdbot.discordcontext import DiscordContext
from mcdbot.logging_integration import integrate_logging

import asyncio
from datetime import datetime
import discord
# from discord.ext import tasks
# import logging
from loguru import logger
import signal


integrate_logging()
global_config = Config()


class Mcdbot(discord.Client):
    def __init__(self, config: Config = global_config):
        global global_config
        global_config = config

        # self.discord_logger = logging.getLogger('discord')
        # self.discord_logger.setLevel(logging.DEBUG if config.debug else logging.WARNING)
        # handler = logging.FileHandler(filename='discord.log', encoding='utf-8', mode='w')
        # handler.setFormatter(logging.Formatter('%(asctime)s:%(levelname)s:%(name)s: %(message)s'))
        # self.discord_logger.addHandler(handler)
        logger.remove()
        self.logger_id = logger.add(config.mcdbot_log,
                                    level="DEBUG" if config.debug else "WARNING",
                                    backtrace=True,
                                    diagnose=config.debug,
                                    catch=False)

        import mcdbot.commands
        from mcdbot.mojang_api import MojangApi
        from mcdbot.rcon import Rcon
        from mcdbot.redis import Redis
        from mcdbot.whitelist import Whitelist

        self.redis = Redis()
        self.rcon = Rcon()
        self.mojang_api = MojangApi()
        self.whitelist = Whitelist(self)

        intents = discord.Intents.none()
        intents.messages = True
        intents.bans = True
        intents.guilds = True
        self.commands = mcdbot.commands.commands
        super().__init__(
            intents=intents,
        )

    def run(self):
        try:
            self.loop.add_signal_handler(signal.SIGUSR1, self.turn_debug_on)
            self.loop.add_signal_handler(signal.SIGINT, self.loop.stop)
            self.loop.add_signal_handler(signal.SIGTERM, self.loop.stop)
        except NotImplementedError:
            pass

        token = global_config.api_token

        async def runner():
            nonlocal self, token
            try:
                logger.info("Starting the main loop...")

                try:
                    await self.redis.start()
                except ConnectionError as ce:
                    logger.error(f"Redis connection error => {ce}")

                try:
                    await self.rcon.start()
                except ConnectionError as ce:
                    logger.error(f"RCON connection error => {ce}")

                try:
                    await self.mojang_api.start()
                except ConnectionError as ce:
                    logger.error(f"Mojang API connection error => {ce}")

                logger.debug(f"Starting with token: {token}")
                await self.start(token)
            except Exception as e:
                logger.exception("Unhandled runner exception", e)
                raise e
            finally:
                logger.warning("Shutting down...")
                await self.mojang_api.stop()
                await self.rcon.stop()
                await self.redis.stop()
                if not self.is_closed():
                    await self.close()

        def stop_loop_on_completion(*_, **__):
            nonlocal self
            self.loop.stop()

        future = asyncio.ensure_future(runner(), loop=self.loop)
        future.add_done_callback(stop_loop_on_completion)

        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            logger.warning("Received KeyboardInterrupt! Shutting down...")
        finally:
            future.remove_done_callback(stop_loop_on_completion)
            discord.client._cleanup_loop(self.loop)

    def turn_debug_on(self):
        # self.discord_logger.setLevel(logging.DEBUG)
        logger.remove(self.logger_id)
        logger.add(global_config.mcdbot_log, level="DEBUG", backtrace=True, diagnose=True, catch=False)

    def get_main_guild(self):
        return self.get_guild(global_config.main_guild_id)

    # @tasks.loop(minutes=global_config.chore_loop_time)
    # async def do_chores(self):
    #     pass
    #
    # @do_chores.before_loop
    # async def wait_before_doing_chores(self):
    #     logger.info('Waiting for the bot to connect before doing chores...')

    async def on_error(self, event_method, *args, **kwargs):
        now = datetime.now().isoformat()
        logger.critical("Uncaught exception!!! Trying to send a message to the discord main channel...")
        logger.critical(f"EXCEPTION INFO: TIME = {now}; event_method = {event_method};"
                        f" args = {args}; kwargs = {kwargs}")
        try:
            admin_role_mention = discord.utils.get(self.get_main_guild().roles,
                                                   id=global_config.admin_role_id).mention
            main_channel = self.get_channel(global_config.main_channel_id)
            await main_channel.send(
                f"{admin_role_mention} "
                f"[CRITICAL] Uncaught exception!!! "
                f"TIME = {now}"
            )
        except Exception as e:
            logger.error("It seems that I can't send the message noticing of the exception. Hmm, that's bad...")
            raise e
        # raise

    async def on_message(self, message):
        try:
            if not message.author.bot:
                if message.content.startswith("!mc "):
                    if len(message.content) >= 150:
                        await message.channel.send(
                            f"{message.author.mention} [ERROR] I don't accept messages longer that 150 characters!")
                        return
                    if isinstance(message.channel, discord.TextChannel):
                        context = DiscordContext.GUILD_TEXT_CHANNEL
                        if message.channel.id != global_config.main_channel_id:
                            await message.channel.send(
                                f"{message.author.mention} [ERROR] This text channel is not my main channel!")
                            return
                    elif isinstance(message.channel, discord.DMChannel):
                        context = DiscordContext.DM_CHANNEL
                    elif isinstance(message.channel, discord.GroupChannel):
                        context = DiscordContext.GROUP_CHANNEL
                    else:
                        raise NotImplementedError("discordpy gave us a channel we don't implement")

                    split = list(filter(lambda i: i != '', message.content.split(" ")))
                    split.pop(0)
                    cmd = split.pop(0)

                    try:
                        cmd_out = await self.commands[context][cmd].run(self, message, context, split)
                        if cmd_out is not None:
                            await message.channel.send(
                                f"{message.author.mention} "
                                f"{cmd_out}"
                            )
                    except KeyError:
                        await message.channel.send(
                            f"{message.author.mention} "
                            f"[ERROR] I don't know a command named `{cmd}`. Try `!mc help` for help.")
                        return
                    except McdbotError as e:
                        await message.channel.send(
                            f"{message.author.mention} [ERROR] {e.NAME}"
                        )
                        return
        except Exception as e:
            logger.critical("We have an uncaught exception!!! Trying to answer the message!")
            logger.exception(e)
            await message.channel.send(
                f"{message.author.mention} [CRITICAL] 500 Internal Bot Error :O"
            )
            raise e



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.mcdbot import Mcdbot
from mcdbot.errors.object_errors import BadMcPlayerError, McPlayerAlreadyRegisteredError, McPlayerNameExistsError
from mcdbot.errors.password_errors import McPasswordInvalidUnderPolicyError
from discord import Member, User
from typing import Union
from loguru import logger


def check_user(main: Mcdbot, user: Union[Member, User]):
    logger.debug(f"Checking {user} of type {type(user)}...")
    if isinstance(user, Member):
        logger.debug(f"Member {user} is a member.")
        return True

    member = main.get_main_guild().get_member(user.id)
    logger.debug(f"User {user} is Member {member}.")
    if member is not None:
        return True
    return False


async def check_mc_player_available(main: Mcdbot, user: Union[Member, User], player: str):
    check_mc_player_name_correctness(player)
    if await main.redis.mc_player_exists(player):
        raise McPlayerNameExistsError
    if await main.redis.user_has_mc_player(user):
        raise McPlayerAlreadyRegisteredError


def check_mc_player_name_correctness(player: str):
    player = player.replace('_', 'Q')
    if len(player) < 3 or len(player) > 16 or not (player.isalnum() and player.isascii() and ' ' not in player):
        raise BadMcPlayerError


def check_mc_password_validity_under_policy(password: str):
    content_policy = (password.isascii() and password.isprintable() and ' ' not in password)
    if len(password) < 6 or len(password) > 30 or not content_policy:
        raise McPasswordInvalidUnderPolicyError



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from discord import Message

from mcdbot.errors.http_like_errors import ForbiddenError
from mcdbot.helpers import check_user, check_mc_player_available, check_mc_password_validity_under_policy
from mcdbot.mcdbot import Mcdbot
from mcdbot.redis import McStatus
from mcdbot.uuid import format_uuid, get_offline_player_uuid


async def register(main: Mcdbot, msg: Message, context, player: str, password: str = None):
    if check_user(main, msg.author):
        await check_mc_player_available(main, msg.author, player)
        if password is None:
            status = McStatus.ONLINE
            profile = await main.mojang_api.get_profile(player)
            uuid = format_uuid(profile['id'])
            player = profile['name']
        else:
            check_mc_password_validity_under_policy(password)
            status = McStatus.OFFLINE
            uuid = get_offline_player_uuid(player)
            await main.rcon.authme_register(player, password)

        await main.rcon.whitelist_add(uuid)

        if not await main.redis.user_set_mc_player(msg.author, player):
            raise RuntimeError(f"user = {msg.author}, player = {player}")
        if not await main.redis.mc_player_set(player, uuid, status, msg.author):
            raise RuntimeError(f"user = {msg.author}, player = {player}")
        else:
            return f"[OK] Registered player '{player}'" \
                   f"({'online' if status == McStatus.ONLINE else 'offline'} account)."
    else:
        raise ForbiddenError



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.mcdbot import global_config
import aioredis
from enum import Enum
from loguru import logger
from discord import Member, User
from typing import Union


class UserStatus(Enum):
    OK = 0
    LOCKED = 1


class McStatus(Enum):
    ONLINE = 0
    OFFLINE = 1


class Redis(object):
    def __init__(self):
        self.started = False
        self._redis = None

    async def start(self):
        if not self.started:
            try:
                self._redis = await aioredis.create_redis_pool(global_config.redis_url)
            except Exception as e:
                self.started = False
                logger.exception("Failed to start the Redis connector!", e)
                raise ConnectionError
            else:
                self.started = True

    async def stop(self):
        if self.started:
            self.started = False
            self._redis.close()
            await self._redis.wait_closed()

    # DB KEYS DEFINITION ###
    @staticmethod
    def _key_user(user: Union[Member, User]):
        return f"user:{user.id}"

    @staticmethod
    def _key_mc_player(mc_player: str):
        return f"mc_player:{mc_player.lower()}"

    # USER ###
    async def user_get_status(self, user: Union[Member, User]):
        await self.start()
        return await self._redis.hget(self._key_user(user), "status")

    async def user_get_mc_player(self, user: Union[Member, User]):
        await self.start()
        return await self._redis.hget(self._key_user(user), "mc_player")

    async def user_set_mc_player(self, user: Union[Member, User], mc_player: str):
        await self.start()
        return await self._redis.hset(self._key_user(user), 'mc_player', mc_player)

    async def user_rm_mc_player(self, user: Union[Member, User]):
        await self.start()
        return await self._redis.hdel(self._key_user(user), 'mc_player')

    async def user_has_mc_player(self, user: Union[Member, User]):
        await self.start()
        return await self._redis.hexists(self._key_user(user), 'mc_player')

    # MC_PLAYER ###
    async def mc_player_exists(self, mc_player: str):
        await self.start()
        return await self._redis.exists(self._key_mc_player(mc_player))

    async def mc_player_set(self, mc_player: str, uuid: str, status: McStatus, owner: Union[Member, User]):
        await self.start()
        return await self._redis.hmset(self._key_mc_player(mc_player),
                                       'name', mc_player,
                                       'uuid', uuid,
                                       'status', status.value,
                                       'owner', owner.id)

    async def mc_player_rm(self, mc_player: str):
        await self.start()
        return await self._redis.delete(self._key_mc_player(mc_player))

    # ATOMIC ###
    async def atomic_purge_mc_player(self, mc_player: str, user: Union[Member, User]):
        """Atomically purges mc_player from db.

        Atomic equivalent of:
        ```
        await redis.mc_player_rm(mc_player)
        await redis.user_rm_mc_player(user)
        ```
        """

        await self.start()

        tr = self._redis.multi_exec()
        tr.delete(self._key_mc_player(mc_player))
        tr.hdel(self._key_user(user), 'mc_player')

        await tr.execute()



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========
import sys

from mcdbot import __version__, __source__, __issues__
from datetime import datetime


class Final(object):
    def __init__(self, **kwargs):
        self.__data = kwargs

    def __getattr__(self, item):
        return self.__data[item]


class Config(object):
    def __init__(self,
                 debug: bool = False,
                 rcon_address='localhost:25575',
                 rcon_password=None,
                 redis_url='redis://localhost',
                 mcdbot_log=None,
                 whitelist_json='whitelist.json',
                 # discordpy_log=None,
                 main_guild_id=None,
                 main_channel_id=None,
                 admin_role_id=None,
                 api_token=None,
                 chore_loop_time=5,
                 ):
        now = datetime.now()

        self.debug = debug
        self.rcon_address = rcon_address
        self.rcon_password = rcon_password
        self.redis_url = redis_url
        self.whitelist_json = whitelist_json
        self.main_guild_id = main_guild_id
        self.main_channel_id = main_channel_id
        self.admin_role_id = admin_role_id
        self.api_token = api_token
        self.chore_loop_time = chore_loop_time

        self.final = Final(
            version=__version__,
            source=__source__,
            issues=__issues__
        )

        self.mcdbot_log = mcdbot_log
        if self.mcdbot_log is None:
            self.mcdbot_log = f"log-{now.isoformat(timespec='seconds')}-mcdbot.log"
        elif self.mcdbot_log == '[stdout]':
            self.mcdbot_log = sys.stdout

        # self.discordpy_log = discordpy_log
        # if self.discordpy_log is None:
        #     self.discordpy_log = f"log-{now.isoformat(timespec='seconds')}-discordpy.log"


default_config_obj = {
    "version": 1,
    "debug": False,
    "mcdbot_log": '[stdout]',
    "whitelist_json": 'whitelist.json',
    "rcon_address": 'localhost:25575',
    "rcon_password": 'RCON PASSWORD HERE',
    "redis_url": 'redis://localhost/0',
    "main_guild_id": 123123,
    "main_channel_id": 321321,
    "admin_role_id": 123123321321,
    "api_token": "BOT API TOKEN HERE",
    "chore_loop_time": 5
}



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.mcdbot import global_config, Mcdbot
import json
from aiofile import AIOFile


class Whitelist(object):
    def __init__(self, main: Mcdbot):
        self.main = main
        try:
            with open(global_config.whitelist_json) as f:
                self.current_whitelist = json.load(f)
        except FileNotFoundError:
            self.current_whitelist = []

    async def _update_whitelist(self):
        async with AIOFile(global_config.whitelist_json, 'w') as f:
            await f.write(json.dumps(self.current_whitelist))
            await f.fsync()
        await self.main.rcon.whitelist_reload()

    async def add_player(self, name: str, uuid: str):
        self.current_whitelist.append({'name': name, 'uuid': uuid})
        await self._update_whitelist()

    async def rm_player(self, name: str):
        deleted = False
        for i in range(len(self.current_whitelist)):
            if self.current_whitelist[i]['name'] == name:
                del self.current_whitelist[i]
                deleted = True
                break
        if deleted:
            await self._update_whitelist()
        return deleted



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.mcdbot import global_config
from mcdbot.discordcontext import DiscordContext
from loguru import logger


async def help_text(main, msg, context):
    from mcdbot.commands import usage_text

    usage = usage_text[context]

    if context == DiscordContext.GUILD_TEXT_CHANNEL:
        context = 'main text channel'
    elif context == DiscordContext.DM_CHANNEL:
        context = 'DMs'
    elif context == DiscordContext.GROUP_CHANNEL:
        context = 'group DMs'
    else:
        logger.warning("[PROBABLY BUG] Context text for context in help_text(msg) is not defined.")
        context = "SOMETHING'S WRONG BRUH"

    return f"""
Mcdbot (v{global_config.final.version})

```
Usage (in the {context}):
{usage}
```

Source code: {global_config.final.source}
Issue tracker: {global_config.final.issues}
"""



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.errors.mojang import MojangProfileNotFoundError, MojangApiHttpStatusError
import aiohttp
import asyncio
from loguru import logger


class MojangApi(object):
    def __init__(self):
        self._session = None

    async def start(self):
        timeout = aiohttp.ClientTimeout(total=30)
        self._session = aiohttp.ClientSession(timeout=timeout)

    async def stop(self):
        await self._session.close()
        await asyncio.sleep(0.25)

    async def get_profile(self, player: str):
        url = f"https://api.mojang.com/users/profiles/minecraft/{player}"
        res = await self._session.get(url)
        if res.status == 204:
            raise MojangProfileNotFoundError
        if res.status == 200:
            return await res.json()
        logger.warning(f"Mojang's API returned unexpected status code: {res.status}")
        raise MojangApiHttpStatusError()



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========


async def unavailiable_in_context(main, msg, context, *args):
    return "[ERROR] The command you used isn't usable in this context!"



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.errors.http_like_errors import ForbiddenError
from mcdbot.errors.object_errors import McPlayerNotFoundError
from mcdbot.helpers import check_user
from mcdbot.mcdbot import Mcdbot


async def unregister(main: Mcdbot, msg, context, player: str):
    if check_user(main, msg.author):
        if main.redis.user_get_mc_player(msg.author) == player:
            await main.redis.atomic_purge_mc_player(player, msg.author)
            await main.rcon.authme_unregister(player)
            await main.whitelist.rm_player(player)
        else:
            raise McPlayerNotFoundError
    else:
        raise ForbiddenError



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from .mcdbot_error import McdbotError


class McPlayerAlreadyRegisteredError(McdbotError):
    NAME = "You have a Minecraft player already registered."


class McPlayerNameExistsError(McdbotError):
    NAME = "This Minecraft player name is already registered - contact the administrator in case of any questions."


class BadMcPlayerError(McdbotError):
    NAME = "Incorrect Minecraft player name."


class McPlayerNotFoundError(McdbotError):
    NAME = "Minecraft player was not found. Beware that Minecraft player names are CASE-SENSITIVE."



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from enum import Enum
from inspect import signature, Parameter
from typing import Callable


class Many(object):
    def __lt__(self, other):
        return False


class ParamDictKeyEnum(Enum):
    NAME = 0
    ANNOTATION = 1
    REQUIRED = 2


class Command(object):
    def __init__(self, handler: Callable, help_text: str = None):
        self.handler = handler
        self.help_text = help_text

        self.params = []
        self.required = 0
        self.max = 0

        sig = signature(handler)
        params = list(sig.parameters.values())
        params.pop(0)
        params.pop(0)
        params.pop(0)

        stop = False
        for i in params:
            if i.kind == i.POSITIONAL_OR_KEYWORD:
                self.max += 1
                required = i.default == Parameter.empty
                if required:
                    if stop:
                        raise ValueError
                    self.required += 1
                else:
                    stop = True
                self.params.append({ParamDictKeyEnum.NAME: i.name,
                                    ParamDictKeyEnum.ANNOTATION: i.annotation,
                                    ParamDictKeyEnum.REQUIRED: required})
            elif i.kind == i.VAR_POSITIONAL:
                self.max = Many()
                self.params.append({ParamDictKeyEnum.NAME: i.name,
                                    ParamDictKeyEnum.ANNOTATION: Many,
                                    ParamDictKeyEnum.REQUIRED: False})
            else:
                raise ValueError

    async def run(self, main, msg, context, split):
        if len(split) < self.required:
            return "[ERROR] You haven't provided all required arguments! Try `!mc help` for help."
        if len(split) > self.max:
            return "[ERROR] You've provided too many arguments! Try `!mc help` for help."
        out = []

        n = len(self.params)
        if isinstance(self.max, Many):
            n -= 1
        for i in range(n):
            param_cast = split.pop(0)
            if self.params[i][ParamDictKeyEnum.ANNOTATION] is str:
                pass
            elif self.params[i][ParamDictKeyEnum.ANNOTATION] is int:
                try:
                    param_cast = int(param_cast)
                except ValueError:
                    return f"[ERROR] The argument `{param_cast}` cannot be converted to a number!" \
                           f" Try `!mc help` for help."
            else:
                raise NotImplementedError
            out.append(param_cast)
        out.extend(split)
        return await self.handler(main, msg, context, *out)



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.mcdbot import Mcdbot


async def start(main: Mcdbot, msg, context):
    dm_channel = msg.author.dm_channel
    if dm_channel is None:
        dm_channel = await msg.author.create_dm()

    await dm_channel.send("Hello! Type !mc help for help.")
    return None



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot import __version__, __source__, __issues__
from mcdbot.config import Config, default_config_obj
from mcdbot.mcdbot import Mcdbot
import json
from loguru import logger
import os
import sys


SUPPORTED_CONFIG_VERSIONS = [1]


if __name__ == '__main__':
    try:
        cmd = sys.argv[1]
    except IndexError:
        cmd = ''

    try:
        path = sys.argv[2]
        os.chdir(os.path.dirname(path))
    except IndexError:
        path = 'mcdbot_config.json'

    if cmd == 'make_or_update_config':
        config = default_config_obj
        try:
            with open(path, 'r') as f:
                config.update(json.load(f))
        except FileNotFoundError:
            pass
        with open(path, 'w') as f:
            json.dump(config, f)
        print("Please configure me in the newly created file.")
    elif cmd == 'run':
        with open(path, 'r') as f:
            config = json.load(f)
        try:
            config_version = config['version']
        except KeyError:
            config_version = None

        if config_version not in SUPPORTED_CONFIG_VERSIONS:
            print("Incompatible config file! Please make a new config file or update your current config file with "
                  "make_or_update_config command.")
            sys.exit(2)

        del config['version']
        config = Config(**config)
        logger.info("Starting up...")
        Mcdbot(config).run()
    elif cmd == 'help':
        print(f"""Mcdbot (v. {__version__})

Usage: python -m mcdbot <COMMAND> [CONFIG_PATH]

Commands:
  make_or_update_config    makes or updates a configuration file at CONFIG_PATH
  run                      starts the bot with configuration from file at CONFIG_PATH
  help                     shows this help text and exits

CONFIG_PATH is 'mcdbot_config.json' by default.
Paths in the config file are relative to the config file.

Source code: {__source__}
Issue tracker: {__issues__}
""")
    else:
        print("Unknown command. Use the command 'help' for help.")
        sys.exit(1)



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from .help_text import help_text
from .register import register
from .unregister import unregister
from .unavailiable_in_context import unavailiable_in_context
from .start import start
from mcdbot.command import Command, ParamDictKeyEnum, Many
from mcdbot.discordcontext import DiscordContext
from loguru import logger

commands = {
    DiscordContext.GUILD_TEXT_CHANNEL: {
        'register': Command(unavailiable_in_context),
        'unregister': Command(unavailiable_in_context),
        'help': Command(help_text, help_text="Show this help."),
        'start': Command(start, help_text="Makes me sneak into your DMs."),
    },
    DiscordContext.DM_CHANNEL: {
        'register': Command(register,
                            "Register your Minecraft player. Type password ONLY if you have a non-official account. DO "
                            "NOT USE YOUR NORMAL PASSWORDS - THEY ARE TRANSMITTED IN PLAIN TEXT THROUGH DISCORD AND "
                            "THROUGH THE INTERNET WHEN LOGGING IN THE MINECRAFT SERVER!"),
        'unregister': Command(unregister,
                              "Unregister your Minecraft player. Type the player name as a confirmation."),
        'help': Command(help_text, help_text="Show this help."),
    },
    DiscordContext.GROUP_CHANNEL: {
        'register': Command(unavailiable_in_context),
        'unregister': Command(unavailiable_in_context),
        'help': Command(help_text, help_text="Show this help."),
        'start': Command(start, help_text="Makes me sneak into your DMs."),
    }
}

usage_text = {}

for context in commands.keys():
    usage_text[context] = []
    max_length = 0
    for name, cmd in commands[context].items():
        if cmd.help_text is not None:
            params = []
            for param in cmd.params:
                if param[ParamDictKeyEnum.ANNOTATION] is str:
                    annotation = ''
                elif param[ParamDictKeyEnum.ANNOTATION] is int:
                    annotation = '=>NUMBER'
                elif param[ParamDictKeyEnum.ANNOTATION] is Many:
                    annotation = '...'
                else:
                    annotation = f'=>{param[ParamDictKeyEnum.ANNOTATION]}'
                    logger.warning("[POSSIBLE BUG] Undefined string for an ANNOTATION, either you're using some "
                                   "plugins or something's wrong...")
                    logger.debug(f"Annotation in question: {param[ParamDictKeyEnum.ANNOTATION]} from {param}")
                title = f"{param[ParamDictKeyEnum.NAME]}{annotation}"
                if param[ParamDictKeyEnum.REQUIRED]:
                    params.append(f"<{title}>")
                else:
                    params.append(f"[{title}]")
            cmd_full = f"{name} {' '.join(params)}"
            max_length = max(max_length, len(cmd_full))
            usage_text[context].append((cmd_full, cmd.help_text))
    text = ''
    for i in usage_text[context]:
        # FIXME: number of spacing dots here
        text += '  ' + i[0].ljust(20, '.') + i[1] + '\n'
    usage_text[context] = text



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

import logging
from loguru import logger


def integrate_logging():
    # Taken from Loguru documentation
    # (https://loguru.readthedocs.io/en/stable/overview.html#entirely-compatible-with-standard-logging)
    class InterceptHandler(logging.Handler):
        def emit(self, record):
            # Get corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where originated the logged message
            frame, depth = logging.currentframe(), 2
            while frame.f_code.co_filename == logging.__file__:
                frame = frame.f_back
                depth += 1

            logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())

    # noinspection PyArgumentList
    logging.basicConfig(handlers=[InterceptHandler()], level=0)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mcdbot', 'mcdbot.commands', 'mcdbot.errors']

package_data = \
{'': ['*']}

install_requires = \
['aiofile>=3.1.1,<4.0.0',
 'aioredis>=1.3.1,<2.0.0',
 'asyncrcon>=1.1.4,<2.0.0',
 'discord.py>=1.5.0,<2.0.0',
 'loguru>=0.5.3,<0.6.0']

setup_kwargs = {
    'name': 'mcdbot',
    'version': '0.3.10',
    'description': 'My PERSONAL project for connecting Discord communities with their Minecraft servers - WIP! NO DOCUMENTATION!',
    'long_description': "# Mcdbot\n\nHello,\n\nThis is my personal project for connecting Discord communities with their Minecraft servers.\n\nI'm making this mainly for myself, and so there is no documentation about the features or installation or anything\nreally.\n\nIf you have any questions, please feel free to create an issue in the issue tracker.\n\n--jsmetana, 2020-10-17T17:18:35+02:00\n",
    'author': 'Jakub Smetana',
    'author_email': 'jakub@smetana.ml',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/jsmetana/mcdbot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#  Copyright (C) 2020 Jakub Smetana <jakub/AT/smetana/DOT/ml>
#  =========
#  SPDX-License-Identifier: MPL-2.0
#  ---------
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#  =========

from mcdbot.mcdbot import global_config
from asyncrcon import AsyncRCON
from loguru import logger


class Rcon(object):
    def __init__(self):
        self.started = False
        self._rcon = AsyncRCON(global_config.rcon_address, global_config.rcon_password)

    async def start(self):
        if not self.started:
            try:
                await self._rcon.open_connection()
            except Exception as e:
                self.started = False
                logger.exception("Failed to start the RCON driver!", e)
                raise ConnectionError
            else:
                self.started = True

    async def stop(self):
        if self.started:
            self.started = False
            self._rcon.close()

    async def authme_register(self, player, password):
        await self.start()
        cmd = f"/authme register {player} {password}"
        out = await self._rcon.command(cmd)
        logger.debug(f"/authme register {player} **** => '''{out}'''")

    async def authme_unregister(self, player):
        await self.start()
        cmd = f"/authme unregister {player}"
        out = await self._rcon.command(cmd)
        logger.debug(f"/authme register {player} **** => '''{out}'''")

    async def authme_password(self, player, password):
        await self.start()
        cmd = f"/authme password {player} {password}"
        out = await self._rcon.command(cmd)
        logger.debug(f"/authme password {player} **** => '''{out}'''")

    async def authme_lastlogin(self, player):
        await self.start()
        cmd = f"/authme lastlogin {player}"
        out = await self._rcon.command(cmd)
        logger.debug(f"/authme lastlogin {player} => '''{out}'''")

    async def authme_accounts(self, player):
        await self.start()
        cmd = f"/authme accounts {player}"
        out = await self._rcon.command(cmd)
        logger.debug(f"/authme accounts {player} => '''{out}'''")

    async def save_off(self):
        await self.start()
        cmd = "/save-off"
        out = await self._rcon.command(cmd)
        logger.debug(f"/save-off => '''{out}'''")

    async def save_on(self):
        await self.start()
        cmd = "/save-on"
        out = await self._rcon.command(cmd)
        logger.debug(f"/save-on => '''{out}'''")

    async def save_all_flush(self):
        await self.start()
        cmd = "/save-all flush"
        out = await self._rcon.command(cmd)
        logger.debug(f"/save-all flush => '''{out}'''")

    # async def whitelist_add(self, entity: str):
    #     await self.start()
    #     cmd = f"/whitelist add {entity}"
    #     out = await self._rcon.command(cmd)
    #     logger.debug(f"/whitelist add {entity} => '''{out}'''")
    #
    # async def whitelist_remove(self, entity: str):
    #     await self.start()
    #     cmd = f"/whitelist remove {entity}"
    #     out = await self._rcon.command(cmd)
    #     logger.debug(f"/whitelist remove {entity} => '''{out}'''")

    async def whitelist_reload(self):
        await self.start()
        cmd = "/whitelist reload"
        out = await self._rcon.command(cmd)
        logger.debug(f"/whitelist reload => '''{out}'''")
